/* $RCSfile: tapmon.c,v $
 *   
 * Copyright (c) 1996 Arizona Board of Regents
 *
 */


#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <signal.h>
#include <ctype.h>
#include <assert.h>
#include <linux/ethtap.h>

#define VERSION 	"1.0"	/* Version numver of process 		*/

#define HTAP_BUFLEN	0x2000	/* 8k buffer 				*/
#define LTAP_BUFLEN	0x2000	/* 8k buffer 				*/

typedef unsigned char	byte_t;

typedef struct {
  int	echo_read	:1;
  int	echo_write	:1;
  int	writes		:1;
  int	mode;
} flags_t;
  
/* 
 * prototypes 
 */

extern void	HexDump(FILE *, const char *, byte_t *, int);

static int	process(int ltap, flags_t ltap_flags,
			int htap,  flags_t htap_flags,
			unsigned long usec);
static void	print_pkts(FILE *, const char *, byte_t *, int);
static void	handler(int signal);

				/* The following are absent from the
                                   header files. */
const char	*basename(const char *);


/* 
 * globals 
 */

static int	shutdown = 0;	/* predicate to determine when to exit  */


/* 
 * code 
 */

int
main(int argc, char *argv[])
{
  const char 	    *usage_command = basename(argv[0]);
  static const char *usage_name    = "tapmon";
  static const char *usage_version = VERSION " (" __DATE__ " " __TIME__ ")";
  static const char *usage_text =
				 "[-dh] [-m minor] [-n usec]"
				 " [-t seconds] [-HL {aprw}]";
  static const char *usage_help =
				 "\tminor\tnet device minor number\n"
				 "\t{apPrw}\ta - active mode\n"
				 "\t\tp - passive mode\n"
				 "\t\tP - passive mode (with writes)\n"
				 "\t\tr - echo reads\n"
				 "\t\tw - echo writes\n"
				 "\t-H\tset htap mode\n"
				 "\t-L\tset ltap mode\n"
				 "\t-h\thelp\n"
				 "\t-m\tspecify minor number (*default = 0*)\n"
				 "\t-n\tspecify micro-sec nap when idle\n"
				 "\t-t\tset run time\n"
				 ;


  int		minor = 0;	/* minor device number 			*/

  flags_t		ltap_flags;
  flags_t		htap_flags;
  long			runtime = -1;
  unsigned long		usecnap = 10;
  int			c;
  int			ltap  = -1; /* /dev/ltap fd 			*/
  int			htap  = -1; /* /dev/htapt fd 			*/
  struct sigaction	sa;	/* POSIX signals 			*/
  char			name[32];
      
				/* set default values 			*/
  ltap_flags.echo_read	 = 0;
  ltap_flags.echo_write = 0;
  ltap_flags.writes	 = 0;
  ltap_flags.mode	 = ETHTAP_PASSIVE_MODE;
  htap_flags.echo_read	 = 0;
  htap_flags.echo_write	 = 0;
  htap_flags.writes	 = 0;
  htap_flags.mode	 = ETHTAP_PASSIVE_MODE;

  opterr = 0;
  while ( (c=getopt(argc,argv,"hm:H:L:t:n:")) != EOF ) {
    switch ( c ) {
      case 't':			/* timeout 				*/
	runtime = atoi(optarg);
	break;
      case 'H':			/* set htap flags 			*/
	if ( strchr(optarg, 'a') ) {
	  ltap_flags.writes = 1;
	  htap_flags.mode = ETHTAP_ACTIVE_MODE;
	}
	if ( strchr(optarg, 'p') )
	  htap_flags.mode = ETHTAP_PASSIVE_MODE;
	if ( strchr(optarg, 'P') ) {
	  ltap_flags.writes = 1;
	  htap_flags.mode   = ETHTAP_PASSIVE_MODE;
	}
	if ( strchr(optarg, 'r') )
	  htap_flags.echo_read = 1;
	if ( strchr(optarg, 'w') )
	  htap_flags.echo_write = 1;
	break;
      case 'L':			/* set ltap flags 			*/
	if ( strchr(optarg, 'a') ) {
	  htap_flags.writes = 1;
	  ltap_flags.mode = ETHTAP_ACTIVE_MODE;
	}
	if ( strchr(optarg, 'p') )
	  ltap_flags.mode = ETHTAP_PASSIVE_MODE;
	if ( strchr(optarg, 'P') ) {
	  htap_flags.writes = 1;
	  ltap_flags.mode   = ETHTAP_PASSIVE_MODE;
	}
	if ( strchr(optarg, 'r') )
	  ltap_flags.echo_read = 1;
	if ( strchr(optarg, 'w') )
	  ltap_flags.echo_write = 1;
	break;
      case 'm':			/* minor device number 			*/
	minor = atoi(optarg);
	break;
      case 'n':			/* usec nap 				*/
	usecnap = atol(optarg);
	break;
      case 'h':			/* help 				*/
	fprintf(stdout, "%s - version %s\n", usage_name, usage_version);
	fprintf(stdout, "Usage: %s %s\n", usage_command, usage_text);
	fprintf(stdout, "%s\n", usage_help);
	exit(0);
      default:			/* usage 				*/
	fprintf(stderr, "%s - version %s\n", usage_name, usage_version);
	fprintf(stderr, "Usage: %s %s\n", usage_command, usage_text);
	exit(-1);
    }
  }
  argc -= optind;
  argv += optind;

  if ( runtime > 0 ) {
    alarm(runtime);
  }

  if ( minor ) 
    sprintf(name, "/dev/ltap%x", minor);
  else
    sprintf(name, "/dev/ltap");
  if ( (ltap = open(name, O_RDWR|O_NONBLOCK)) == -1 ) {
    char	buf[128];

    sprintf(buf, "cannot open %s", name);
    perror(buf);
    exit(-1);
  }
  if ( minor )
    sprintf(name, "/dev/htap%x", minor);
  else
    sprintf(name, "/dev/htap");
  if ( (htap = open(name, O_RDWR|O_NONBLOCK)) == -1 ) {
    char	buf[128];
    
    sprintf(buf, "cannot open %s", name);	
    perror(buf);
    exit(-1);
  }
      
  sa.sa_handler = handler;
  sa.sa_flags   = 0;
  sigemptyset(&sa.sa_mask);
  sigaction(SIGINT, &sa, NULL);
  sigaction(SIGALRM, &sa, NULL);
 
  process(ltap, ltap_flags, htap, htap_flags, usecnap);

  close(ltap);
  close(htap);

  return 0;
} /* main */


static void
handler(int signal) 
{
  switch ( signal ) {
    case SIGINT:
      shutdown = 1;
      break;
      /* Once the timer expires try nicely to stop the process.  If
	 the process does not exit after 5 seconds exit. */
    case SIGALRM:
      printf("alarm expired\n");
      if ( shutdown )
	exit(-1);		/* bail out! 				*/
      shutdown = 1;
      alarm(5);
      break;
    default:
      ;
  }
} /* handler */



static int
process(int ltap, flags_t ltap_flags, int htap, flags_t htap_flags,
	unsigned long usec)
{
  byte_t	ltap_buffer[LTAP_BUFLEN];
  byte_t	htap_buffer[HTAP_BUFLEN];
  int		x;
  
  assert(ltap != -1);
  assert(htap != -1);

				/* Set active or passive mode 		*/
  if ( htap_flags.mode == ETHTAP_ACTIVE_MODE ) {
    if ( ioctl(htap, ETHTAPSETMODE, &htap_flags.mode) )
      perror("htap ioctl ETHTAPSETMODE failed");
  }
  if ( ltap_flags.mode == ETHTAP_ACTIVE_MODE ) {
    if ( ioctl(ltap, ETHTAPSETMODE, &ltap_flags.mode) )
      perror("ltap ioctl ETHTAPSETMODE failed");
  }

				/* Verify active of passive mode 	*/
  if ( ioctl(ltap, ETHTAPGETMODE, &x) )
    perror("ltap ioctl ETHTAPGETMODE failed");
  else
    printf("ltap mode is %s\n",
	   x == ETHTAP_PASSIVE_MODE ? "PASSIVE" : "ACTIVE");

  if ( ioctl(htap, ETHTAPGETMODE, &x) )
    perror("htap ioctl ETHTAPGETMODE failed");
  else
    printf("htap mode is %s\n",
	   x == ETHTAP_PASSIVE_MODE ? "PASSIVE" : "ACTIVE");

  ioctl(htap, ETHTAPZEROCOUNTS, NULL);
  ioctl(ltap, ETHTAPZEROCOUNTS, NULL);
      
  while ( !shutdown ) {
    int		ltap_read = 0;
    int		htap_read = 0;
    int		active    = 0;	/* activity level 			*/

    ltap_read = read(ltap, ltap_buffer, sizeof ltap_buffer);
    htap_read = read(htap, htap_buffer, sizeof htap_buffer);

    if ( ltap_read > 0 ) {
      if ( ltap_flags.echo_read ) {
	HexDump(stdout, "ltap[r]", ltap_buffer, ltap_read);
	puts("");
      }
      active++;
    }
    if ( htap_read > 0 ) {
      if ( htap_flags.echo_read ) {
	print_pkts(stdout, "htap[r] ", htap_buffer, htap_read);
      }
      active++;
    }

	    /* At this point we could send the ltap and htap traffic
	       through the x-kernel.  For now just write the read
	       buffers back to the devices so NET traffic continues.
	       If we have a passive connection the writes do nothing
	       so don't worrying about doubling traffic. */

    if ( ltap_read > 0 && htap_flags.writes ) {
      if ( htap_flags.echo_write ) {
	HexDump(stdout, "htap[w]", ltap_buffer, ltap_read);
	puts("");
      }
      write(htap, ltap_buffer, ltap_read);
    }

    if ( htap_read > 1 && ltap_flags.writes ) {
	if ( ltap_flags.echo_write ) {
	  print_pkts(stdout, "ltap[w] ", htap_buffer, htap_read);
	}
	write(ltap, htap_buffer, htap_read);
    }
    
				/* no activity so go to sleep for a
                                   while */
    if ( usec && !active ) {
      usleep(usec);
    }
    
  }
  
  return 0;
} /* process */



static void
print_pkts(FILE *fout, const char *label, byte_t *buf, int count)
{
  unsigned char		pkts = *(unsigned char *)buf;
  unsigned short	len;
  byte_t		*p = buf + sizeof pkts;
  int			i;

  for (i=0; i < pkts; i++) {
    len = *(unsigned short *)p;
    p  += sizeof len;
    HexDump(fout, label, p, len);
    p += len;
    puts("");
  }
/*   printf("\n%s: ------\n", label); */
  fflush(fout);
} /* print_pkts */





