/*
 * $RCSfile: dhkx.c,v $
 *
 * Diffie-Hellman Key Exchange protocol
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993  Arizona Board of Regents
 *
 * 
 * $Log: dhkx.c,v $
 * Revision 1.6  1994/10/31 18:23:36  menze
 * Fixed #endif comment
 *
 * Revision 1.5  1994/10/01  01:29:49  ho
 * Rom option "ownhlptype" can be used so that dhkx is not "inline"
 * and will use its own type for hlpType in actions on its transport
 * protocol.
 *
 * Revision 1.4  1994/07/01  00:17:54  rcs
 * Merged in code for Dubner board.
 *
 * Revision 1.3  1994/06/28  23:53:28  ho
 * Added key manager interface, options for one or two participants, rom opts
 *
 * Revision 1.2  1994/02/05  00:05:48  menze
 *   [ 1994/02/03          menze ]
 *   Added casts, fixed prototypes
 *
 */

/* Diffie-Hellman Key Exchange protocol:

   P and g are public.  P is a large prime.
   Alice chooses random X (mod P-1) and communicates g^X (mod P) to Bob.
   Bob chooses random Y (mod P-1) and communicates g^Y to Alice.
   Each then computes g^XY and selects 56 bits to use as a key.
*/


/* Open issues, some marked in the code with a $$.

   Fix GMP library on the 386.
   Does GMP leave any intermediate data on the stack?  Probably.
     Consider installing some erasures in the GMP code.
   Consider being more aggressive about erasing intermediate values, and
     storing less in the session state.  Maybe only x and g^x really needs
     to be remembered during the setup?
   The flag gmp_memory_setup needs to be declared once/xkernel.  How to arrange?
   Several bignum functions should really be once/xkernel.  How to arrange?
   setbignum and arrayfrombignum could be much more efficient by taking
     advantage of the internal representation of bignums.  Not used enough
     to matter; also the present code is portable.
   Provide for releasing the xk master lock while computing.
   Byte order note:  Only Big Endian testing so far.
   xkrandom:  Drop it for now, but leave hooks in.
   No provision for defending against the "man in the middle" attack.
     This requires adding signatures.
   Perhaps dhkxClose should invalidate & erase the keys in the KMGR sessions.
   In the interest of speed, dhkxKeyExchangeReply has a timing glitch: it
     calculates & sends the reply message before calculating & setting the key.
     Another message might arrive before we've set the key; it will be dropped.
 */

/* DHKX misc design notes:

   RVRs mods to event code require "returning" semaphores.  I'm ignoring this.

   I decided not to do retries for the present.  If the offer message is lost,
   the code will time out after ten seconds and the Open will return failure.
   If I were to do retries, I should remember recent offers, in case an old
   one is accepted.

   If an Open is attempted on a session that is being set up, it fails.
   Arguably it should be queued.

   OpenDone at higher level can't send any messages until KX has processed
   first popped message with KX offer.

   Certain kinds of surprize Close are impossible because noone has a
   pointer to the current session.

   Future versions might provide for use of primes > 512 bits.
*/


/* #define TIME */  /* Turns on timing code.  Prints times to do key exchange. */
/* #define DUBNER */  /* Dubner card additions; replaces GNU bignum code. */

#ifdef DUBNER
#include "rcrunch.h"
#include "rhfuncs.h"
#endif


#include "xkernel.h"
#include "dhkx.h"
#include "dhkx_i.h"
#include "gmp.h"
#include "eth.h"		/* So ARP can go on top of DHKX */
#include "ip.h"                 /* For IP pseudoheader fixup. */
#include "km.h"

extern int tracedhkxp; int tracedhkxp;


#ifdef __STDC__

static XObj		dhkxCreateSessn( XObj, XObj, XObj, XObj);
static void		dhkxSessnInit( XObj );
static int		dhkxControlProtl( XObj, int, char *, int );
static xkern_return_t	dhkxDemux( XObj, XObj, Msg * );
static XObj		dhkxOpen( XObj, XObj, XObj, Part * );
static xkern_return_t	dhkxOpenEnable( XObj, XObj, XObj, Part * );
static xkern_return_t	dhkxOpenDisable( XObj, XObj, XObj, Part * );
static xkern_return_t	dhkxOpenDone( XObj, XObj, XObj, XObj);
static xkern_return_t	dhkxClose( XObj );
static xmsg_handle_t	dhkxPush( XObj, Msg * );
static xkern_return_t	dhkxPop( XObj, XObj, Msg *, void * );
static int		dhkxControlSessn( XObj, int, char *, int );
static void		zap_MP_INT(MP_INT *);
static void		setbignum(MP_INT *, unsigned long *, long);
static void		arrayfrombignum(unsigned long *, MP_INT *, long);
static void		dhkxKeyExchangeInitiate(XObj);
static xkern_return_t	dhkxKeyExchangeReply(XObj, Msg *);
static xkern_return_t	dhkxKeyExchangeDone(XObj, Msg *);
static void		dhkxOpenWakeup(Event, void *);
/* int			xkrandom(unsigned char *, long); */
static long		dhkxcopy(void *, char *, long, void *);
static void		xkFreeZ(void *, size_t);
static void *		xkReallocZ(void *, size_t, size_t);

#else

static XObj		dhkxCreateSessn();
static void		dhkxSessnInit();
static int		dhkxControlProtl();
static xkern_return_t	dhkxDemux();
static XObj		dhkxOpen();
static xkern_return_t	dhkxOpenEnable();
static xkern_return_t	dhkxOpenDisable();
static xkern_return_t	dhkxOpenDone();
static xkern_return_t	dhkxClose();
static xmsg_handle_t	dhkxPush();
static xkern_return_t	dhkxPop();
static int		dhkxControlSessn();
static void		zap_MP_INT();
static void		setbignum();
static void		arrayfrombignum();
static void		dhkxKeyExchangeInitiate();
static xkern_return_t	dhkxKeyExchangeReply();
static xkern_return_t	dhkxKeyExchangeDone();
static void		dhkxOpenWakeup();
/* int			xkrandom(); */
static long		dhkxcopy();
static void		xkFreeZ();
static void *		xkReallocZ();

#endif /* __STDC__ */

#ifdef TIME
static void subtime(t1, t2, t3)    XTime *t1, *t2, *t3;
{   t3->sec = t2->sec - t1->sec;
    t3->usec = t2->usec - t1->usec;
    if (t3->usec < 0) { t3->usec += 1000000; t3->sec -= 1; } }
#endif

/* Base prime: Smallest solution of P = 2Q+1 = 8K+3, P & Q prime, P > 2^511.5.
   2 is a QNR of P = 8K+3, therefore a primitive root (since P=2Q+1).
   Old value of dhkxmodstr ended with 05519, used 7 as primitive root, and
   omitted the P = 8K+3 requirement. */
char *dhkxmodstr = "948075190810917672683252645565215926008454174403132986379"
                   "244333505065230347814082479545572840742073300693309061417"
                   "9782624068317238241310650437075740662819";
MP_INT dhkxmod,dhkxbase;


/* These three functions are the interface between the xkernel memory allocation
   functions and the GMP allocation requirements.  I've arranged for the
   free and realloc operations to zero out returned memory, for security. */

extern int gmp_memory_setup;  /* One-time flag for mp_set_memory_functions. */
int gmp_memory_setup = 0;  /* $$ Figure out a way to get this declared once. */

/* As long as xMalloc is a function, this define is ok. */
#define xkgmpMalloc xMalloc

static void xkFreeZ(ptr,size) void *ptr; size_t size;
{ xTrace2(dhkxp,TR_FULL_TRACE,"xkFreeZ: ptr %x size %d",ptr,size);
  bzero(ptr,size); xFree(ptr); }

static void * xkReallocZ(ptr,oldsize,newsize) void *ptr; size_t oldsize,newsize;
{  void * val;
   xTrace3(dhkxp,TR_FULL_TRACE,"xkReallocZ: ptr %x oldsize %d newsize %d",
	   ptr,oldsize,newsize);
   val = xkgmpMalloc(newsize);
   if (oldsize >= newsize)
   { printf("xkReallocZ: oldsize %d > newsize %d\n", oldsize, newsize);
     Kabort("xkReallocZ"); }
   bcopy(ptr,val,oldsize);
   xkFreeZ(ptr,oldsize);
   xTrace1(dhkxp,TR_FULL_TRACE,"xkReallocZ returns val %x", val);
   return val; }

/* Clears out bignum, then zeros lengths & pointer fields. */
static void zap_MP_INT(mpint) MP_INT* mpint;
{  xTrace1(dhkxp,TR_FULL_TRACE,"zap_MP_INT: mpint %x", mpint);
   mpz_clear(mpint); bzero((void *)mpint,sizeof(MP_INT));  }

/* Copy a value from an array to a bignum.  Takes care of possible leading 0s.
   Assumes the bignum contains a value, and clears it before overwriting. */
static void setbignum(pbn,arr,l) MP_INT *pbn; unsigned long *arr; long l;
{ xTrace3(dhkxp,TR_FULL_TRACE,"setbignum: pbn %x arr %x l %d", pbn, arr, l);
  mpz_set_ui(pbn,0);
  for (;l>0;arr++,l--) { mpz_mul_2exp(pbn,pbn,32); mpz_add_ui(pbn,pbn,*arr); } }

/* Copy a value from a bignum into an array.  Arrange for the higher order
   part of the bignum to go into the lower part of the array, so this is
   inverse to setbignum.  Doesn't care if there's bignum left over. */
static void arrayfrombignum(arr,pbn,l) MP_INT *pbn; unsigned long *arr; long l;
{ xTrace3(dhkxp,TR_FULL_TRACE,"arrayfrombignum: arr %x pbn %x l %d", arr, pbn, l);
  for (l--;l>=0;l--) { arr[l] = mpz_get_ui(pbn); mpz_div_2exp(pbn,pbn,32); } }


#ifdef DUBNER

#define dubmod av
#define dubg bv
#define dubx cv
#define dubgx dv
#define duby tempv4
#define dubgy tempv5
#define dubgxy tempv6
#define dubgxtmp tempv6 /* shared with gxy */

/* Format of cruncher number:  Unsigned only; a block of bytes; low-to-high. */

static void array_to_cruncher(arr,crreg,l) unsigned long *arr; hreg* crreg; long l;
{  crreg->size = 4*l; crreg->normbits = 0; crreg->flags = 0;
   for (;l>0;l--,arr++)
     { put_left(crreg->location+4*l-4, *arr & 255);
       put_left(crreg->location+4*l-3, (*arr>>8) & 255);
       put_left(crreg->location+4*l-2, (*arr>>16) & 255);
       put_left(crreg->location+4*l-1, (*arr>>24) & 255); };
   hshorten(crreg);
   xIfTrace(dhkxp, TR_DETAILED) { hdump(crreg); };
  }

/* Copies at most L words; zero filled if necessary; not an error if
   there's leftover data; sign of crreg ignored, just use the magnitude. */
static void cruncher_to_array(crreg,arr,l) unsigned long *arr; hreg* crreg; long l;
{  unsigned long a,b,c,d, *arrt = arr, zbytes; long ll = l;
   xIfTrace(dhkxp, TR_DETAILED) { hdump(crreg); };
   for (;l>0;l--,arrt++)
     { a = get_left(crreg->location+4*l-4);
       b = get_left(crreg->location+4*l-3);
       c = get_left(crreg->location+4*l-2);
       d = get_left(crreg->location+4*l-1);
       *arrt = ((d & 255)<<24) + ((c&255)<<16) + ((b&255)<<8) + (a&255); }
   /* If number is shorter than expected, clear out the "high" bytes of arr.
      this is complicated, since arr is in words, and its high significance is at
      low addresses.  But the last few bytes to zero are at high addresses in
      the partially zeroed word.  */
   if (crreg->size < 4*ll)
     { zbytes = crreg->size - 4*ll;
       if (zbytes >= 4) bzero((unsigned char *)arr, zbytes & ((unsigned long)~3));
       if (zbytes & 3) *(arr+zbytes/4) &= (0xffffff >> (8*(zbytes&3) - 8)); }; }

#endif /* DUBNER */


/* Initiate a key exchange.
   Called from Open, with the session (in state DHKXINIT) as an argument.
   Composes a key exchange initiation message, and ships it.
   Sets session state to DHKXINITWAIT to indicate we are waiting for a reply
   message to complete the exchange. */
static void dhkxKeyExchangeInitiate(sessn) XObj sessn;
{ SState *ss;
  unsigned long xbuf[16], msgbuf[1+1+16]; MP_INT gxtmp; Msg msg;

  xTrace1(dhkxp, TR_MAJOR_EVENTS, "dhkxKeyExchangeInitiate entered: session = %x", sessn);
  xAssert(xIsSession(sessn));
  ss = sessn->state;
  xTrace2(dhkxp, TR_MORE_EVENTS, "dhkxKeyExchangeInitiate: ss = %x, state = %d",
	  ss, ss->state);
  xAssert(ss->state == DHKXINIT);

#ifdef TIME
  xGetTime(&ss->starttime);
#endif

  /* Get 512 random bits. */
  /* $$ Temporarily disconnect xkrandom, till I make it more robust. */
  /* if (xkrandom((unsigned char *)xbuf, 16 * sizeof(long))) printf("xkrandom failed"); */
  { int i; for (i=0;i<16;i++) xbuf[i] = i*i+5; };
  xIfTrace(dhkxp, TR_DETAILED)
  {int i; printf("dhkxKeyExchangeInitiate:\nxbuf =");
   for (i=0;i<16;i++) printf("%s %08x%s",(i==8)?"     ":"",xbuf[i],((i&7)==7)?"\n":""); };

#ifndef DUBNER
  mpz_init(&ss->x); setbignum(&ss->x,xbuf,16);
#else
  array_to_cruncher(xbuf,dubx,16);
#endif
  bzero((char *)xbuf, 16 * sizeof(long));

#ifndef DUBNER
  /* x and gx saved in ss */
  mpz_init(&ss->gx); mpz_powm(&ss->gx,&dhkxbase,&ss->x,&dhkxmod);  /* gx = g^x (mod P) */
#else
  hmove(dubmod,dubgx);
  hpowerm(dubgx,dubg,dubx); /* modulus & result both in dubgx!? */
#endif

  xIfTrace(dhkxp, TR_DETAILED)
  { printf("dhkxKeyExchangeInitiate: g^x = ");
#ifndef DUBNER
    mpz_out_str(stdout,10,&ss->gx);
#else
    fhdispa(stdout,dubgx);
#endif
    printf("\n"); }

  /* build message with DHKXMAGIC, state DHKXINIT, gx padded to 512 bits. */
  msgbuf[0] = DHKXMAGIC; msgbuf[1] = ss->state;
#ifndef DUBNER
  mpz_init_set(&gxtmp,&ss->gx); arrayfrombignum(&msgbuf[2],&gxtmp,16);
#else
  cruncher_to_array(dubgx,&msgbuf[2],16);
#endif
  xIfTrace(dhkxp, TR_DETAILED)
  {int i; printf("dhkxKeyExchangeInitiate: msg is ");
   for (i=0;i<18;i++)
     printf("%s %08x%s",(i==8)?"     ":"",msgbuf[i],(i%10==7)?"\n":""); };
#ifndef DUBNER
  zap_MP_INT(&gxtmp); /* I could just mpz_clear gxtmp, since it's not private. */
#endif

  /* ship it, destroy message */
  msgConstructBuffer(&msg,(char *)msgbuf, (1+1+16) * sizeof(long));
  xPush(xGetDown(sessn,TRANSPORT_PROTOCOL),&msg);
  msgDestroy(&msg);

  ss->state = DHKXINITWAIT;  /* set state to DHKXINITWAIT */
  xTrace1(dhkxp, TR_EVENTS, "dhkxKeyExchangeInitiate exit: state = %d", ss->state);
}


/* Receive a key exchange initiation message, and reply.
   Called from Demux, with the session (in state DHKXRESPONDER) and setup message
   as arguments.
   Computes the session key based on the setup message and our random data;
   compose and send an appropriate reply to the initiator, for him to
   complete his side of the key setup.
   Enter the computed key in the KM sessions.
   Set session state to DHKXTRANSPARENT, putting us in transparent mode. */
static xkern_return_t dhkxKeyExchangeReply(sessn,msg) XObj sessn; Msg *msg;
{ SState *ss;
  PState *ps = sessn->myprotl->state;
  unsigned long ybuf[16], msgbuf[1+1+16+16]; MP_INT gytmp,ktmp; unsigned char key[8];
  int j,t,tmp; Msg rmsg;

  xTrace2(dhkxp, TR_MAJOR_EVENTS,
	  "dhkxKeyExchangeReply entered: session = %x, msg = %x", sessn, msg);
  xAssert(xIsSession(sessn));
  ss = sessn->state;
  xTrace2(dhkxp, TR_MORE_EVENTS, "dhkxKeyExchangeReply: ss = %x, state = %d",
	  ss, ss->state);
  xAssert(ss->state == DHKXRESPONDER);

#ifdef TIME
  xGetTime(&ss->starttime);
#endif

  /* Check message length; unpack it. */
  xTrace1(dhkxp, TR_DETAILED, "dhkxKeyExchangeReply: msgLen = %d", msgLen(msg));
  if ((tmp = msgLen(msg)) != sizeof(long) * (1+1+16))
  { xTrace4(dhkxp, TR_GROSS_EVENTS,
	"dhkxKeyExchangeReply: Session %x ss %x: Offer message %x length %d is wrong",
	    sessn, ss, msg, tmp);
    return XK_FAILURE; };
  msgPop(msg, dhkxcopy, (char *) msgbuf, (1+1+16)*sizeof(long), 0);

  xIfTrace(dhkxp, TR_DETAILED)
  {int i; printf("dhkxKeyExchangeReply: msg is ");
   for (i=0;i<18;i++)
     printf("%s %08x%s",(i==8)?"     ":"",msgbuf[i],(i%10==7)?"\n":""); };

  /* Check msg starts with DHKXMAGIC, and state of sender is DHKXINIT. */
  if (!(msgbuf[0] == DHKXMAGIC && msgbuf[1] == DHKXINIT))
  { xTrace5(dhkxp, TR_GROSS_EVENTS,
	    "dhkxKeyExchangeReply: Session %x ss %x: Offer message %x."
	    " Magic %d or SendState %d is wrong",
	    sessn, ss, msg, msgbuf[0], msgbuf[1]);
    return XK_FAILURE; };

  /* 512 bits -> gx */
#ifndef DUBNER
  mpz_init(&ss->gx); setbignum(&ss->gx,&msgbuf[2],16);
#else
  array_to_cruncher(&msgbuf[2],dubgx,16);
#endif
  xIfTrace(dhkxp, TR_DETAILED)
  { printf("dhkxKeyExchangeReply: g^x = ");
#ifndef DUBNER
    mpz_out_str(stdout,10,&ss->gx);
#else
    fhdispa(stdout,dubgx);
#endif
    printf("\n"); }

#ifndef DUBNER
  if (mpz_cmp(&ss->gx,&dhkxmod) >= 0)
#else
  if (hcomp(dubgx,dubmod) >= 0)
#endif
  { xTrace3(dhkxp, TR_GROSS_EVENTS,
	    "dhkxKeyExchangeReply: Session %x ss %x: Offer message %x."
	    " g^x isn't < modulus\n",
	    sessn, ss, msg);
    xIfTrace(dhkxp, TR_GROSS_EVENTS) {
#ifndef DUBNER
      mpz_out_str(stdout,10,&ss->gx);
#else
      fhdispa(stdout,dubgx);
#endif
      printf("\n"); }
    return XK_FAILURE; }

  /* Call xkrandom to get y, 512 bits. */
  /* $$ xkrandom is temporarily turned off. */
  /* if (xkrandom((unsigned char *)ybuf, 16 * sizeof(long))) printf("xkrandom failed"); */
  { int i; for (i=0;i<16;i++) ybuf[i] = i*i*i+105; };
  xIfTrace(dhkxp, TR_DETAILED)
  {int i; printf("dhkxKeyExchangeReply:\nybuf =");
   for (i=0;i<16;i++) printf("%s %08x%s",(i==8)?"     ":"",ybuf[i],((i&7)==7)?"\n":""); };
#ifndef DUBNER
  mpz_init(&ss->y); setbignum(&ss->y,ybuf,16);
#else
  array_to_cruncher(ybuf,duby,16);
#endif
  bzero((char *)ybuf, 16 * sizeof(long));

#ifndef DUBNER
  mpz_init(&ss->gy); mpz_powm(&ss->gy,&dhkxbase,&ss->y,&dhkxmod); /* gy = g^y (mod P) */
#else
  hmove(dubmod,dubgy);
  hpowerm(dubgy,dubg,duby); /* modulus & result both in dubgx!? */
#endif
  xIfTrace(dhkxp, TR_DETAILED)
  { printf("dhkxKeyExchangeReply: g^y = ");
#ifndef DUBNER
    mpz_out_str(stdout,10,&ss->gy);
#else
    fhdispa(stdout,dubgy);
#endif
    printf("\n"); }
  ss->state = DHKXREPLYING;  /* Advance session state to indicate we are replying. */

  /* reply message: DHKXMAGIC, state = DHKXREPLYING, gx & gy each padded to 512 */
  /* DHKXMAGIC & gx are already present in msgbuf */
  msgbuf[1] = ss->state;
#ifndef DUBNER
  mpz_init_set(&gytmp,&ss->gy); arrayfrombignum(&msgbuf[2+16],&gytmp,16);
#else
  cruncher_to_array(dubgy,&msgbuf[2+16],16);
#endif
  xIfTrace(dhkxp, TR_DETAILED)
  {int i; printf("dhkxKeyExchangeReply: reply msg is ");
   for (i=0;i<34;i++)
     printf("%s %08x%s",(i%9==7)?"     ":"",msgbuf[i],(i%9==6)?"\n":""); };
#ifndef DUBNER
  zap_MP_INT(&gytmp); /* I could just mpz_clear gytmp, since it's not private. */
#endif

  /* Ship the reply.
  $$ Timing glitch: xPush releases the master lock.
     I could get another msg before I've finished computing and initialized
     the encryption keys.
     As long as the encryption protocol is above me in the graph, I can just
     throw away messages that arrive before I've finished computing the key.
     Presumably the correct thing to do is queue them. */
  msgConstructBuffer(&rmsg,(char *)msgbuf, (1+1+16+16) * sizeof(long));
  xPush(xGetDown(sessn,TRANSPORT_PROTOCOL),&rmsg);
  msgDestroy(&rmsg);  /* destroy message */

#ifdef TIME
  xGetTime(&ss->endtime);
#endif

  /* set key from gxy */
#ifndef DUBNER
  mpz_init(&ss->gxy); mpz_powm(&ss->gxy,&ss->gx,&ss->y,&dhkxmod); /* gxy = gx^y (mod P) */
#else
  hmove(dubmod,dubgxy);
  hpowerm(dubgxy,dubgx,duby); /* modulus & result both in dubgx!? */
#endif
  xIfTrace(dhkxp, TR_DETAILED)
  { printf("dhkxKeyExchangeReply: g^xy = ");
#ifndef DUBNER
    mpz_out_str(stdout,10,&ss->gxy);
#else
    fhdispa(stdout,dubgxy);
#endif
    printf("\n"); }
#ifndef DUBNER
  mpz_init_set(&ktmp,&ss->gxy); arrayfrombignum((unsigned long *)key,&ktmp,2);
#else
  cruncher_to_array(dubgxy,(unsigned long *)key,2);
#endif
  xIfTrace(dhkxp, TR_DETAILED)
  {int i; printf("dhkxKeyExchangeReply: key = ");
   for (i=0;i<8;i++) printf(" %02x", key[i]); printf("\n"); }
#ifndef DUBNER
  zap_MP_INT(&ktmp); zap_MP_INT(&ss->gxy);
#endif

  /* fix parity on key - each byte should be odd */
  for (j=0;j<8;j++) { t = key[j]; t ^= t>>4; t ^= t>>2; t ^= t>>1; key[j] ^= (~t)&1; }
  xIfTrace(dhkxp, TR_DETAILED)
  {int i; printf("dhkxKeyExchangeReply: odd parity key = ");
   for (i=0;i<8;i++) printf(" %02x", key[i]); printf("\n"); }

  /* Enter the key in two KM sessions, one for each host. */
  if (ps->flags & ALL_PARTS)
  { xTrace1(dhkxp, TR_DETAILED, "dhkxKeyExchangeReply: key session is %x",
	    xGetDown(sessn,KEY_MGR_SESS_ALL)); }
  else
  { xTrace2(dhkxp, TR_DETAILED, "dhkxKeyExchangeReply: key sessions are %x and %x",
	  xGetDown(sessn,KEY_MGR_SESS_IN), xGetDown(sessn,KEY_MGR_SESS_OUT)); }
  if (!(ps->flags & ALL_PARTS)) 
    xControl(xGetDown(sessn,KEY_MGR_SESS_IN), KM_SET, (char *)key, 8);
  xControl(xGetDown(sessn,KEY_MGR_SESS_OUT), KM_SET, (char *)key, 8); /* OUT==ALL */
  bzero((char *)key,8); t=0;  /* Zero out key array on stack. */
  /* At this point, I don't really need to remember anything: the key has been
     handed off to the key manager, and all my intermediate values can be trashed. */
#ifndef DUBNER
  zap_MP_INT(&ss->y); /* Actually, I didn't even need to store these in ss. */
#endif

  ss->state = DHKXTRANSPARENT;  /* go transparent */

#ifdef TIME
  xGetTime(&ss->endtime2);
  subtime(&ss->starttime,&ss->endtime,&ss->deltatime);
  subtime(&ss->starttime,&ss->endtime2,&ss->deltatime2);
  printf("\nResponder replied after %d.%06d seconds, and finished key calculation after %d.%06d seconds.\n",
     ss->deltatime.sec, ss->deltatime.usec, ss->deltatime2.sec, ss->deltatime2.usec);
#endif

  xTrace1(dhkxp, TR_EVENTS, "dhkxKeyExchangeReply exit: state = %d", ss->state);
  return XK_SUCCESS;
}



/* Handle the reply to the key exchange initiation message.
   Called from Demux, with the session (in state DHKXINITWAIT) and reply message
   as arguments.
   Check that reply message matches our key initiation message.
   Set session state to DHKXCOMPUTING.
   Computes the session key based on our initiation data and the
   other information in the reply message.
   Enter the computed key in the KM sessions.
   Set session state to DHKXTRANSPARENT, putting us in transparent mode. */
static xkern_return_t dhkxKeyExchangeDone(sessn,msg) XObj sessn; Msg *msg;
{ SState *ss;
  PState *ps = sessn->myprotl->state;
  unsigned long msgbuf[1+1+16+16]; MP_INT gxtmp,ktmp; unsigned char key[8];
  int j,t,tmp;

  xTrace2(dhkxp, TR_MAJOR_EVENTS,
	  "dhkxKeyExchangeDone entered: session = %x, msg = %x", sessn, msg);
  xAssert(xIsSession(sessn));
  ss = sessn->state;
  xTrace2(dhkxp, TR_MORE_EVENTS, "dhkxKeyExchangeDone: ss = %x, state = %d",
	  ss, ss->state);
  xAssert(ss->state == DHKXINITWAIT);

  /* Check message length; unpack it. */
  xTrace1(dhkxp, TR_DETAILED, "dhkxKeyExchangeDone: msgLen = %d", msgLen(msg));
  if ((tmp = msgLen(msg)) != (1+1+16+16)*sizeof(long))
  { xTrace4(dhkxp, TR_GROSS_EVENTS,
	"dhkxKeyExchangeDone: Session %x ss %x: Reply message %x length %d is wrong",
	    sessn, ss, msg, tmp);
    return XK_FAILURE; };
  msgPop(msg, dhkxcopy, (char *) msgbuf, (1+1+16+16)*sizeof(long), 0);

  xIfTrace(dhkxp, TR_DETAILED)
  {int i; printf("dhkxKeyExchangeDone: msg is ");
   for (i=0;i<34;i++)
     printf("%s %08x%s",(i%9==7)?"     ":"",msgbuf[i],(i%9==6)?"\n":""); };

  /* Check msg starts with DHKXMAGIC, and state of sender is DHKXREPLYING. */
  if (!(msgbuf[0] == DHKXMAGIC && msgbuf[1] == DHKXREPLYING))
  { xTrace5(dhkxp, TR_GROSS_EVENTS,
	    "dhkxKeyExchangeDone: Session %x ss %x: Reply message %x."
	    " Magic %d or SendState %d is wrong",
	    sessn, ss, msg, msgbuf[0], msgbuf[1]);
    return XK_FAILURE; };

  /* 512 bits -> gxtmp; Make sure it's same as gx that we sent. */
#ifndef DUBNER
  mpz_init(&gxtmp); setbignum(&gxtmp,&msgbuf[2],16);
#else
  array_to_cruncher(&msgbuf[2],dubgxtmp,16);
#endif
  xIfTrace(dhkxp, TR_DETAILED)
  { printf("dhkxKeyExchangeDone: reply g^x = ");
#ifndef DUBNER
    mpz_out_str(stdout,10,&gxtmp);
#else
    fhdispa(stdout,dubgxtmp);
#endif
    printf("\n");
    printf("dhkxKeyExchangeDone: offer g^x = ");
#ifndef DUBNER
    mpz_out_str(stdout,10,&ss->gx);
#else
    fhdispa(stdout,dubgx);
#endif
    printf("\n"); }
#ifndef DUBNER
  if (mpz_cmp(&gxtmp,&ss->gx))
#else
  if (hcomp(dubgxtmp,dubgx))
#endif
  { xIfTrace(dhkxp, TR_GROSS_EVENTS)
    { printf("dhkxKeyExchangeDone: Session %x ss %x: Reply message %x."
	     "  Reply g^x doesn't match offer g^x.\n",
	     sessn, ss, msg);
#ifndef DUBNER
      mpz_out_str(stdout,10,&gxtmp); printf("\n");
      mpz_out_str(stdout,10,&ss->gx); printf("\n");
#else
      fhdispa(stdout,dubgxtmp); printf("\n");
      fhdispa(stdout,dubgx); printf("\n");
#endif
    }
    return XK_FAILURE; };

  /* 512 bits -> gy. */
#ifndef DUBNER
  mpz_init(&ss->gy); setbignum(&ss->gy,&msgbuf[2+16],16);
#else
  array_to_cruncher(&msgbuf[2+16],dubgy,16);
#endif
  xIfTrace(dhkxp, TR_DETAILED)
  { printf("dhkxKeyExchangeDone: g^y = ");
#ifndef DUBNER
    mpz_out_str(stdout,10,&ss->gy);
#else
    fhdispa(stdout,dubgy);
#endif
    printf("\n"); }

  /* Verify that gy < dhkxmod. */
#ifndef DUBNER
  if (mpz_cmp(&ss->gy,&dhkxmod) >= 0)
#else
  if (hcomp(dubgy,dubmod) >= 0)
#endif
  { xTrace3(dhkxp, TR_GROSS_EVENTS,
	    "dhkxKeyExchangeReply: Session %x ss %x: Reply message %x."
	    " Reply g^y isn't < modulus\n",
	    sessn, ss, msg);
    xIfTrace(dhkxp, TR_GROSS_EVENTS) {
#ifndef DUBNER
      mpz_out_str(stdout,10,&ss->gy);
#else
      fhdispa(stdout,dubgx);
#endif
      printf("\n"); }
    return XK_FAILURE; }

  ss->state = DHKXCOMPUTING;  /* Now committed to the Open. */

  /* set key from gxy */
#ifndef DUBNER
  mpz_init(&ss->gxy); mpz_powm(&ss->gxy,&ss->gy,&ss->x,&dhkxmod); /* gxy = gy^x (mod P) */
#else
  hmove(dubmod,dubgxy);
  hpowerm(dubgxy,dubgy,dubx); /* modulus & result both in dubgxy!? */
#endif
  xIfTrace(dhkxp, TR_DETAILED)
  { printf("dhkxKeyExchangeDone: g^xy = ");
#ifndef DUBNER
    mpz_out_str(stdout,10,&ss->gxy);
#else
    fhdispa(stdout,dubgxy);
#endif
    printf("\n"); }
#ifndef DUBNER
  mpz_init_set(&ktmp,&ss->gxy); arrayfrombignum((unsigned long *)key,&ktmp,2);
#else
  cruncher_to_array(dubgxy,(unsigned long *)key,2);
#endif
  xIfTrace(dhkxp, TR_DETAILED)
  {int i; printf("dhkxKeyExchangeDone: key = ");
   for (i=0;i<8;i++) printf(" %02x", key[i]); printf("\n"); }
#ifndef DUBNER
  zap_MP_INT(&ktmp); zap_MP_INT(&ss->gxy);
#endif

  /* fix parity on key - each byte should be odd */
  for (j=0;j<8;j++) { t = key[j]; t ^= t>>4; t ^= t>>2; t ^= t>>1; key[j] ^= (~t)&1; }
  xIfTrace(dhkxp, TR_DETAILED)
  {int i; printf("dhkxKeyExchangeDone: odd parity key = ");
   for (i=0;i<8;i++) printf(" %02x", key[i]); printf("\n"); }

  /* Enter the key in two KM sessions, one for each host. */
  if (ps->flags & ALL_PARTS)
  { xTrace1(dhkxp, TR_DETAILED, "dhkxKeyExchangeDone: key session is %x",
	    xGetDown(sessn,KEY_MGR_SESS_ALL)); }
  else
  { xTrace2(dhkxp, TR_DETAILED, "dhkxKeyExchangeDone: key sessions are %x and %x",
	  xGetDown(sessn,KEY_MGR_SESS_IN), xGetDown(sessn,KEY_MGR_SESS_OUT)); }
  if (!(ps->flags & ALL_PARTS))
    xControl(xGetDown(sessn,KEY_MGR_SESS_IN), KM_SET, (char *)key, 8);
  xControl(xGetDown(sessn,KEY_MGR_SESS_OUT), KM_SET, (char *)key, 8); /* OUT==ALL */
  bzero((char *)key,8); t=0;  /* Zero out key array on stack. */

  /* There's a timing problem here if the encryption module is below us in the
   graph: there might be an immediate 2nd message awaiting us.  This is unlikely,
   but might happen as a "one-shot" RPC reply.  [If our offer message were bundled
   with an RPC, or an implied RPC.  We don't presently do this.]  In fact, the
   2nd message might even have raced ahead of the KX reply, and been discarded.
   This is all theory right now; but it's a good idea to keep the encryption module
   is above us in the graph.
  */

  /* At this point, I don't really need to remember anything: the key has been
     handed off to the key manager, and all my intermediate values can be trashed. */
#ifndef DUBNER
  zap_MP_INT(&ss->y);  /* Actually, I didn't need to store ss->gxy, etc. in ss. */
#endif
  /*  I don't need to erase message, since it's assumed to be tapped. */

  ss->state = DHKXTRANSPARENT;  /* go transparent */

#ifdef TIME
  xGetTime(&ss->endtime);
  subtime(&ss->starttime,&ss->endtime,&ss->deltatime);
  printf("\nKey exchange completed after %d.%06d seconds.\n",
	 ss->deltatime.sec, ss->deltatime.usec);
#endif

  xTrace1(dhkxp, TR_EVENTS, "dhkxKeyExchangeDone exit: state = %d", ss->state);
  return XK_SUCCESS;
}


static xkern_return_t romfinddelay(XObj, char **, int, int, void *);
static xkern_return_t romfindpartnum(XObj, char **, int, int, void *);
static xkern_return_t romfindhlptype(XObj, char **, int, int, void *);

static XObjRomOpt rmopts[3] = {{ "delay", 1, romfinddelay },
			       { "numparts", 1, romfindpartnum},
			       { "ownhlptype", 0, romfindhlptype}
			       };

static xkern_return_t
romfinddelay(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  ((PState *)(self->state))->delay = atoi(ptr[2]);
  return XK_SUCCESS;
}

static xkern_return_t
romfindhlptype(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  ((PState *)(self->state))->ownhlptype = 1;
  return XK_SUCCESS;
}

static xkern_return_t
romfindpartnum(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  if (atoi(ptr[2]) > 1)
    ((PState *)(self->state))->flags |= ALL_PARTS;
  else
    ((PState *)(self->state))->flags &= ~ALL_PARTS;
  return XK_SUCCESS;
}

void
dhkx_init( self )
     XObj self;
{
    PState	*ps;
    XObj	llp;

    xTrace1(dhkxp, TR_ERRORS, "dhkx_init entered: protocol = %x", self);
    xAssert(xIsProtocol(self));
    
#ifdef DUBNER
    main2();
#endif

    if ( ! xIsProtocol(llp = xGetDown(self, TRANSPORT_PROTOCOL)) )
    { xError("dhkx_init: dhkx cannot get lower protocol object"); return; }

    ps = X_NEW(PState);
    self->state = (VOID *)ps;
    ps->actMap = mapCreate(DHKX_ACTIVE_MAP_SZ, sizeof(ActiveId));
    ps->pasMap = mapCreate(DHKX_PASSIVE_MAP_SZ, sizeof(PassiveId));
    ps->delay = DHKX_TIMEOUT;
    ps->flags = 0;
    ps->ownhlptype = 0;

    if (!gmp_memory_setup)
    { xTrace0(dhkxp,TR_EVENTS,"dhkx: setting up GMP memory allocation functions");
      mp_set_memory_functions((void *(*)(size_t))xkgmpMalloc,xkReallocZ,xkFreeZ);
      gmp_memory_setup = 1; }

#ifndef DUBNER
    mpz_init_set_str(&dhkxmod,dhkxmodstr,10);
    mpz_init_set_ui(&dhkxbase,2);
#else
    hstring(dhkxmodstr,dubmod);
    hsetval(2,dubg);
#endif

    self->control = dhkxControlProtl;
    self->open = dhkxOpen;
    self->openenable = dhkxOpenEnable;
    self->opendisable = dhkxOpenDisable;
    self->opendone = dhkxOpenDone;
    self->close = dhkxClose;  
    self->demux = dhkxDemux;
    findXObjRomOpts(self, rmopts, 3, 0);
    xTrace0(dhkxp, TR_EVENTS, "dhkx_init exiting");
}

/* Wake up dhkxOpen: Either the session timed out, or the reply arrived.
   evnt is unused, so that Demux can call the same wakeup function. */
static void
dhkxOpenWakeup(evnt, session) Event evnt; void *session;
{ SState	*ss;
  xTrace2(dhkxp,TR_EVENTS,"dhkxOpenWakeup called; evnt = %x, session = %x", evnt, session);
  xAssert(xIsSession((XObj)session));
  ss = ((XObj)session)->state;
  evCancel(ss->event);  /* cancel timeout event - might be our thread, doesn't matter */
  ss->event = (Event) 0;
  semSignal(&ss->semaphore);
 }

static XObj
dhkxOpen( self, hlpRcv, hlpType, part )
    XObj	self, hlpRcv, hlpType;
    Part 	*part;
{
    XObj 	dhkxSessn;
    XObj	lls;
    PState	*pstate;
    SState	*ss;

    xTrace4(dhkxp,TR_EVENTS,"dhkxOpen entered with args %x %x %x %x",
	    self,hlpRcv,hlpType,part);

    pstate = (PState *) self->state;
    
    if ( part == 0 || partLen(part) < 1 )
  { xTrace0(dhkxp, TR_SOFT_ERRORS, "dhkxOpen -- bad participants");   return ERR_XOBJ; }

    xTrace4(dhkxp,TR_EVENTS,"dhkxOpen calling xOpen with args %x %x %x %x",
	    self,hlpType,xGetDown(self,TRANSPORT_PROTOCOL),part);

    lls = xOpen(self, pstate->ownhlptype?self:hlpType,
		xGetDown(self, TRANSPORT_PROTOCOL), part);
    if ( lls == ERR_XOBJ )
    { xTrace0(dhkxp, TR_ERRORS, "dhkxOpen: could not open lls"); return ERR_XOBJ; }

    if ( mapResolve(pstate->actMap, &lls, &dhkxSessn) == XK_SUCCESS )
    { xTrace0(dhkxp, TR_MAJOR_EVENTS, "found an existing one"); xClose(lls); }
    else
    { xTrace0(dhkxp, TR_MAJOR_EVENTS, "creating a new one");
      if ( (dhkxSessn = dhkxCreateSessn(self, hlpRcv, hlpType, lls)) == ERR_XOBJ )
      { xClose(lls);   return ERR_XOBJ; } }

    ss = (SState *)dhkxSessn->state;
    if (ss->state != DHKXNEW && ss->state != DHKXTRANSPARENT)
    { xTrace1(dhkxp, TR_ERRORS,
	      "dhkxOpen: tried to do second Open while first in-progress, state = %d",
	      ss->state);
      return ERR_XOBJ; };

    /* If we got a fresh session, start the key exchange protocol.
       Send off an offer message. */
    if (ss->state == DHKXNEW)
    { ss->state = DHKXINIT;
      xDuplicate(dhkxSessn); /* Bump the reference count on the session, since
	     it will have message traffic prior to returning from dhkxOpen. */
      dhkxKeyExchangeInitiate(dhkxSessn);
  loop:
      xTrace2(dhkxp, TR_EVENTS, "dhkxOpen: sleeping for %d seconds; ss = %x",
	      pstate->delay, ss);
      /* schedule time out after DHKX_TIMEOUT seconds */
      ss->event = evSchedule(dhkxOpenWakeup,dhkxSessn,pstate->delay*1000000);
      semInit(&ss->semaphore,0);
      semWait(&ss->semaphore);  /* the pause that refreshes */
      xTrace2(dhkxp, TR_EVENTS, "dhkxOpen: awakened; ss = %x, state = %d",ss,ss->state);
      /* When we get here, either timed out, or got reply. */
      if (ss->state == DHKXCOMPUTING)
      { xTrace1(dhkxp,TR_EVENTS,
		"dhkxOpen: reply received; still computing key.  ss = %x",ss);
	goto loop; }; /* We are willing to wait forever if computing. */
      if (ss->state == DHKXINITWAIT)  /* Timed out.  Clean up session & fail. */
      { xTrace1(dhkxp,TR_ERRORS,"dhkxOpen timed out, returning failure. ss = %x",ss);
	dhkxSessn->rcnt--;  dhkxClose(dhkxSessn);  return ERR_XOBJ; };
      if (ss->state != DHKXTRANSPARENT)  /* Bogus state: bomb. */
      { printf("\ndhkxOpen:  Session state %x has bad state value %d\n", ss, ss->state);
	Kabort("Aborting from dhkxOpen."); };
      /* OK!  Drop through. */
      dhkxSessn->rcnt--;  /* Drop refcnt to 0; reincremented when I exit dhkxOpen. */
    }

    xTrace1(dhkxp,TR_EVENTS,"dhkxOpen returning session %x", dhkxSessn);
    return (dhkxSessn);
}


static XObj
dhkxCreateSessn( self, hlpRcv, hlpType, lls)
    XObj	self, hlpRcv, hlpType, lls;
{
    XObj	s;
    SState	*ss;
    PState	*ps = (PState *)self->state;
    Part        part[5];  /* So I can partInternalize partBuf, even if 5 participants. */
    char        partBuf[100];
    XObj        SessnArray[3];
    XObj        kms1;
    XObj        kms2;
    Part	part1[2];
    Part	part2[1];
    int		tmp;

    xTrace4(dhkxp, TR_EVENTS, "dhkxCreateSessn entered: args %x %x %x %x",
	    self, hlpRcv, hlpType, lls);

    if (xControl(lls, GETPARTICIPANTS, partBuf, 100) <= 0)
    { xTrace0(dhkxp, TR_ERRORS, "dhkxCreateSessn: could not get participants");
      return ERR_XOBJ; }
	
    if ((tmp = partExtLen(partBuf)) > 5)
    { xTrace1(dhkxp, TR_ERRORS,
	      "dhkxCreateSessn: Too many %d participants; max 5.", tmp);
      return ERR_XOBJ; }

    partInternalize(part,partBuf);

    if (partLen(part) < 2)
    { xTrace0(dhkxp, TR_ERRORS, "dhkxCreateSessn: too few participants");
      return ERR_XOBJ; }

    if (!(ps->flags & ALL_PARTS)) /* open two key manager sessions */
    { partInit(part1,1);
      part1[0] = part[REMOTE_PART];
      part1[0].len = 1; /* major hack */
      kms1 = xOpen(self, self, xGetDown(self,KEY_MGR_PROTOCOL), part1);
      if (kms1 == ERR_XOBJ)
      { xTrace0(dhkxp, TR_ERRORS, "dhkxCreateSessn: could not open key manager");
	return ERR_XOBJ; }
      partInit(part2,1);
      part2[0] = part[LOCAL_PART];
      part2[0].len = 1; /* major hack */
      kms2 = xOpen(self, self, xGetDown(self,KEY_MGR_PROTOCOL), part2);
      if (kms2 == ERR_XOBJ)
      { xTrace0(dhkxp, TR_ERRORS, "dhkxCreateSessn: could not open key manager");
	xClose(kms1);
	return ERR_XOBJ; }
    }
    else    /* open one key manager session */
    { partInit(part1,2);
      part1[REMOTE_PART] = part[REMOTE_PART];
      part1[LOCAL_PART] = part[LOCAL_PART];
      part1[REMOTE_PART].len = part1[LOCAL_PART].len = 2;  /* ugly hack */
      kms1 = xOpen(self, self, xGetDown(self,KEY_MGR_PROTOCOL), part1);
      if (kms1 == ERR_XOBJ)
      { xTrace0(dhkxp, TR_ERRORS, "dhkxCreateSessn: could not open key manager");
	return ERR_XOBJ; }
    }

    SessnArray[0]=lls;
    SessnArray[1]=kms1;
    if (!(ps->flags & ALL_PARTS)) SessnArray[2]=kms2;
    s = xCreateSessn(dhkxSessnInit, hlpRcv, hlpType, self,
		     (ps->flags & ALL_PARTS) ? 2 : 3, SessnArray);
    s->binding = (Bind) mapBind(ps->actMap, (char *)&lls, (int)s);

    if ( s->binding == ERR_BIND ) {
	xTrace0(dhkxp, TR_ERRORS, "dhkxCreateSessn: mapBind error");
	xClose(kms1);
 	if (!(ps->flags & ALL_PARTS)) xClose(kms2);
	xDestroy(s);
	return ERR_XOBJ;
      }

    ss = X_NEW(SState);
    xTrace1(dhkxp, TR_FULL_TRACE, "dhkxCreateSessn: X_NEW(SState) returned %x", ss);
    bzero((void *)ss, sizeof(SState));
    ss->state = DHKXNEW;
    s->state = (VOID *)ss;

    xTrace1(dhkxp, TR_MORE_EVENTS, "dhkxCreateSessn returning %x",s);
    return s;
}


static xkern_return_t
dhkxOpenEnable(self, hlpRcv, hlpType, part)
    XObj self, hlpRcv, hlpType;
    Part *part;
{
    PState	*ps = (PState *)self->state;
    xkern_return_t val;

    xTrace4(dhkxp, TR_EVENTS, "dhkxOpenEnable entered: args %x %x %x %x",
	    self, hlpRcv, hlpType, part);

    if ( defaultOpenEnable(ps->pasMap, hlpRcv, hlpType, 
	ps->ownhlptype?(VOID *)&self:(VOID *)&hlpType) == XK_FAILURE )
    { xTrace0(dhkxp, TR_MORE_EVENTS, "dhkxOpenEnable: defaultOpenEnable failed ");
      return XK_FAILURE; }
    
    val = xOpenEnable(self, ps->ownhlptype?self:hlpType, 
		      xGetDown(self,TRANSPORT_PROTOCOL), part);
    if ( val == XK_FAILURE )
    { xTrace0(dhkxp, TR_MORE_EVENTS, "dhkxOpenEnable: xOpenEnable failed ");
      return XK_FAILURE; }

    xTrace1(dhkxp, TR_MORE_EVENTS, "dhkxOpenEnable returning %x", val);
    return val;
} 


static xkern_return_t
dhkxOpenDisable(self, hlpRcv, hlpType, part)
    XObj self, hlpRcv, hlpType;
    Part *part;
{
    PState	*ps = (PState *)self->state;
    xkern_return_t val;

    xTrace4(dhkxp, TR_EVENTS, "dhkxOpenDisable entered: args %x %x %x %x",
	    self, hlpRcv, hlpType, part);

    val = defaultOpenDisable(ps->pasMap, hlpRcv, hlpType, 
			     ps->ownhlptype?(void*)&self:(void*)&hlpType);

    xTrace1(dhkxp, TR_MORE_EVENTS, "dhkxOpenDisable returning %x", val);
    return val;
}

static xkern_return_t
dhkxOpenDone(self, lls, llp, hlpType)
    XObj self, lls, llp, hlpType;
{
    XObj        s;
    PState      *pstate;
    Enable      *e;
   
    xTrace4(dhkxp, TR_MAJOR_EVENTS, "Entered dhkxOpenDone: args %x %x %x %x",
	    self, lls, llp, hlpType);

    if ( self == hlpType ) {
        xTrace0(dhkxp, TR_ERRORS, "self == hlpType in dhkxOpenDone");
      /*  return XK_FAILURE; */
    }
    pstate = (PState *)self->state;

    /*
     * check for open enables
     */
    if ( mapResolve(pstate->pasMap, &hlpType, &e) == XK_FAILURE ) {
        /*
         * This shouldn't happen
         */
        xTrace0(dhkxp, TR_ERRORS,
                "dhkxOpenDone: Couldn't find hlp for incoming session");
        return XK_FAILURE;
    }

    xDuplicate(lls);
    if ( (s = dhkxCreateSessn(self, e->hlpRcv, e->hlpType, lls))
                == ERR_XOBJ ) {
        xTrace0(dhkxp, TR_ERRORS,
                "dhkxOpenDone: Couldn't create session");
        xClose(lls);
        return XK_FAILURE;
    }

    if (((SState *)(s->state))->state == DHKXNEW)
      ((SState *)(s->state))->state = DHKXRESPONDER;
    else   /* Barf on any other state.  Some other path must have tried to use it? */
    { xTrace1(dhkxp, TR_ERRORS,
	      "dhkxOpenDone: Wrong state %d for newly created session.",
	      ((SState *)(s->state))->state);
      xClose(lls);
      return XK_FAILURE; }

    xTrace0(dhkxp, TR_EVENTS,
            "dhkxOpenDone: Passively opened session successfully created");
    return xOpenDone(e->hlpRcv, s, self);
}



static xkern_return_t
dhkxDemux( self, lls, msg )
    XObj	self, lls;
    Msg		*msg;
{
  PState	*ps = (PState *)self->state;
  XObj 	s = 0;
  xkern_return_t val;
  
  xTrace3(dhkxp, TR_EVENTS, "dhkxDemux entered: args %x %x %x",
	  self, lls, msg);

  xIfTrace(dhkxp, TR_DETAILED) msgShow(msg);

  if ( mapResolve(ps->actMap, &lls, &s) == XK_FAILURE ) { 
	xTrace0(dhkxp, TR_EVENTS, "dhkxDemux: map resolve fails");
  	return XK_FAILURE;
  }

  if ( xIsSession(s) )  val = xPop(s, lls, msg, 0);
  else val = XK_FAILURE;

  if (val == XK_SUCCESS)
  xTrace0(dhkxp, TR_FUNCTIONAL_TRACE, "dhkxDemux returning XK_SUCCESS");
  else if (val == XK_FAILURE)
  xTrace0(dhkxp, TR_FUNCTIONAL_TRACE, "dhkxDemux returning XK_FAILURE");
  else
  xTrace1(dhkxp, TR_FUNCTIONAL_TRACE, "dhkxDemux returning %x", val);

  return val;
}


static xkern_return_t
dhkxClose( s )
    XObj	s;
{
    PState	*ps = (PState *)xMyProtl(s)->state;
    xkern_return_t res;
    SState	*ss;

    xTrace1(dhkxp, TR_MAJOR_EVENTS, "dhkxClose: closing session %x", s);
    xAssert( xIsSession( s ) );
    xAssert( s->rcnt == 0 );
    res = mapRemoveBinding( ps->actMap, s->binding );
    xAssert( res != XK_FAILURE );
    /* Clear out SState. */
    if (ss = s->state)
    { /* $$ Security note: it's convenient to keep around MP_INTs for debugging,
        but they should actually be discarded as soon as keys are established. */
      if (ss->x.alloc) mpz_clear(&ss->x);
      if (ss->gx.alloc) mpz_clear(&ss->gx);
      if (ss->y.alloc) mpz_clear(&ss->y);
      if (ss->gy.alloc) mpz_clear(&ss->gy);
      if (ss->gxy.alloc) mpz_clear(&ss->gxy);
      /* I don't have to do anything about semaphore & event, because they
	 are only used while waiting for reply & computing, and nobody will
	 have the session pointer to call Close with. */
      bzero((void *)ss,sizeof(SState)); }   /* The xDestroy will xFree(ss). */
    xClose(xGetDown(s,TRANSPORT_PROTOCOL));
    xClose(xGetDown(s,KEY_MGR_SESS_OUT));  /* OUT==ALL */  /* Perhaps invalidate key before close? $$ */
    if (!(ps->flags & ALL_PARTS)) xClose(xGetDown(s,KEY_MGR_SESS_IN));  /* Perhaps invalidate key before close? $$ */
    xDestroy( s );
    xTrace0(dhkxp, TR_DETAILED, "dhkxClose returning");
    return XK_SUCCESS;
}

static
long dhkxcopy(to,from,len,ignore)
    void *to; char *from; void *ignore; long len;
{
  xTrace4(dhkxp, TR_FULL_TRACE, "dhkxcopy args: %x %x %x %x", to,from,len,ignore);

  bcopy(from,to,len);  return len;
}

static xmsg_handle_t
dhkxPush( s, msg )
    XObj	s;
    Msg 	*msg;
{ int val;

  xTrace2(dhkxp, TR_EVENTS, "dhkxPush: session %x  msg %x", s, msg);

  if (((SState *)(s->state))->state == DHKXTRANSPARENT)
    val = xPush(xGetDown(s, TRANSPORT_PROTOCOL), msg);
  else { xTrace2(dhkxp, TR_ERRORS,
	"dhkxPush: Session isn't in transparent state %d; dropping message %x",
		 ((SState *)(s->state))->state, msg);
	 val = XK_FAILURE; }

  xTrace1(dhkxp, TR_DETAILED, "dhkxPush exiting, value %x", val);

  return val;
}


static xkern_return_t
dhkxPop( s, lls, m, hdr)
    XObj	s, lls;
    Msg 	*m;
    void        *hdr;
{ SState *ss = (SState *)s->state;

  xTrace5(dhkxp, TR_EVENTS, "dhkxPop: args %x %x %x %x.  state %d.",
	  s, lls, m, hdr, ss->state);
  switch (ss->state)
  { case DHKXTRANSPARENT:  /* the ordinary case, after everything is set up */
      return xDemux(s, m);

    case DHKXINITWAIT:     /* process the reply to our offer message */
      if (dhkxKeyExchangeDone(s, m) == XK_SUCCESS)
	{ dhkxOpenWakeup(0,s); return XK_SUCCESS; } /* if ok, wakeup Open */
      else return XK_FAILURE;
      /* Fail if something wrong with reply; Open will eventually time out anyway. */

    case DHKXRESPONDER:    /* reply to an offer message */
      return dhkxKeyExchangeReply(s, m);
      /* Returns XK_FAILURE if offer has bad format. */

    /* Other possible states:
       REPLYING, if the replier is slow computing the key.
       COMPUTING, possibly. */

    default:
      xTrace2(dhkxp, TR_ERRORS,
	      "dhkxPop: Unexpected session state %x, dropping message %x",
	      ss->state, m);
      return XK_FAILURE;  }
}


static int
dhkxControlProtl( self, op, buf, len )
    XObj	self;
    int 	op, len;
    char 	*buf;
{ xTrace4(dhkxp, TR_EVENTS, "dhkxControlProtl: args %x %x %x %x", self, op, buf, len);
  xAssert(xIsProtocol(self));
  return xControl(xGetDown(self, TRANSPORT_PROTOCOL), op, buf, len);
}

static int
dhkxControlSessn(s, op, buf, len)
    XObj	s;
    int 	op, len;
    char 	*buf;
{ xTrace4(dhkxp, TR_EVENTS, "dhkxControlSessn: args %x %x %x %x", s, op, buf, len);
  xAssert(xIsSession(s));
  return xControl(xGetDown(s,TRANSPORT_PROTOCOL), op, buf, len);
}


static void
dhkxSessnInit(s)
    XObj s;
{ xTrace1(dhkxp, TR_EVENTS, "dhkxSessnInit: session %x", s);
  s->push = dhkxPush;
  s->close = dhkxClose;
  s->control = dhkxControlSessn;
  s->pop = dhkxPop;
}

