/*
 * tcp_subr.c
 *
 * Derived from:
 *
 * Copyright (c) 1982, 1986 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of California at Berkeley. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 *
 *	@(#)tcp_subr.c	7.13 (Berkeley) 12/7/87
 *
 * Modified for x-kernel v3.2
 * Modifications Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 * $Revision: 1.1 $
 * $Date: 1996/07/11 21:39:21 $
 */

#include "xkernel.h"
#include "ip.h"
#include "icmp.h"
#include "tcp_internal.h"
#include "tcp_fsm.h"
#include "tcp_seq.h"
#include "tcp_timer.h"
#include "tcp_var.h"
#include "tcpip.h"

int	tcp_s_ttl = TCP_S_TTL;

/*
 * Create template to be used to send tcp_s packets on a connection.
 * Call after host entry created, allocates an mbuf and fills
 * in a skeletal tcp_s/ip header, minimizing the amount of work
 * necessary when the connection is used.
 */
struct tcp_siphdr *
tcp_s_template(tp)
	struct tcp_scb *tp;
{
	register struct inpcb *inp = tp->t_inpcb;
	register struct tcp_siphdr *n;
	XObj	lls, tcp_sSessn;
	IPpseudoHdr	pHdr;

	if ((n = tp->t_template) == 0) {
		n = (struct tcp_siphdr *)xMalloc(sizeof *n);
	}
	tcp_sSessn = tcp_scbtoso(tp);
	xAssert(xIsSession(tcp_sSessn));
	lls = xGetDown(tcp_sSessn,0);
	xAssert(xIsSession(lls));
	if ( xControl(lls, IP_GETPSEUDOHDR, (char *)&pHdr, sizeof(pHdr)) < 0 ) {
	    xTraceP0(tcp_sSessn, TR_ERRORS,
		     "tcp_s_template could not get pseudo-hdr from lls");
	    pHdr.prot = 0;
	}
	n->ti_next = n->ti_prev = 0;
	n->ti_x1 = 0;
	n->ti_pr = pHdr.prot;
	n->ti_len = sizeof (struct tcp_siphdr) - sizeof (struct ipovly);
	n->ti_src = inp->inp_laddr;
	n->ti_dst = inp->inp_raddr;
	n->ti_sport = inp->inp_lport;
	n->ti_dport = inp->inp_rport;
	n->ti_seq = 0;
	n->ti_ack = 0;
	n->ti_x2 = 0;
	n->ti_off = 5;
	n->ti_flags = 0;
	n->ti_win = 0;
	n->ti_sum = 0;
	n->ti_urp = 0;
	/*
	 * IP pseudo-header
	 */
	n->ti_p.src = *(IPhost *)&n->ti_src;
	n->ti_p.dst = *(IPhost *)&n->ti_dst;
	n->ti_p.zero = 0;
	n->ti_p.prot = n->ti_pr;
	return (n);
}

/*
 * Send a single message to the TCP_S at address specified by
 * the given TCP_S/IP header.  If flags==0, then we make a copy
 * of the tcp_siphdr at ti and send directly to the addressed host.
 * This is used to force keep alive messages out using the TCP_S
 * template for a connection tp->t_template.  If flags are given
 * then we send a message back to the TCP_S which originated the
 * segment ti, and discard the mbuf containing it and any other
 * attached mbufs.
 *
 * In any case the ack and sequence number of the transmitted
 * segment are as specified by the parameters.
 */
void
tcp_s_respond(tp, th, pHdr, ack, seq, flags, tcp_s)
	struct tcp_scb *tp;
    	register struct tcp_shdr *th;
    	register IPpseudoHdr *pHdr;
	tcp_s_seq ack, seq;
	int flags;
	XObj tcp_s;
{
	int win = 0, tlen;
	Msg m;
	XObj s = NULL;
	struct tcp_shdr tHdr;
	hdrStore_t	store;

	if (tp) {
		s = tp->t_inpcb->inp_session;
		win = sototcp_sst(s)->rcv_space;
		s = xGetDown(s, 0);
	}
#ifdef TCP_S_COMPAT_42
	tlen = flags == 0;
#else
	tlen = 0;
#endif
	tHdr = *th;
	if (flags == 0) {
		flags = TH_ACK;
	} else {
	    	u_short tmp;

		tmp = tHdr.th_sport;
		tHdr.th_sport = tHdr.th_dport;
		tHdr.th_dport = tmp;
	}
	tHdr.th_seq = seq;
	tHdr.th_ack = ack;
	tHdr.th_x2 = 0;
	tHdr.th_off = sizeof (struct tcp_shdr) >> 2;
	tHdr.th_flags = flags;
	tHdr.th_win = win;
	tHdr.th_urp = 0;
	msgConstructEmpty(&m);
	
	store.m = &m;
	store.h = pHdr;
	msgPush(&m, tcp_sHdrStore, &tHdr, sizeof(tHdr), &store);
	if (s) {
	    xPush(s, &m);
	} else {
	    Part p[2];
	    
	    partInit(p, 2);
	    partPush(p[0], &pHdr->dst, sizeof(IPhost));
	    partPush(p[1], &pHdr->src, sizeof(IPhost));
	    s = xOpen(tcp_s, tcp_s, xGetDown(tcp_s, 0), p);
	    if ( xIsSession(s) ) {
		xPush(s, &m);
		xClose(s);
	    } else {
		xError("tcp_s_respond could not open lower session!");
	    }
	}
	msgDestroy(&m);
}


/*
 * Create a new TCP_S control block, making an
 * empty reassembly queue and hooking it to the argument
 * protocol control block.  Assumes that the ip control block and associated
 * tcp_s session already exist.
 */
struct tcp_scb *
tcp_s_newtcp_scb(inp)
	struct inpcb *inp;
{
	register struct tcp_scb *tp;

	tp = (struct tcp_scb *)xMalloc(sizeof *tp);
	bzero((char*)tp, sizeof *tp);
	tp->seg_next = tp->seg_prev = (struct reass *)tp;
	tp->t_maxseg = TCP_S_MSS;
	tp->t_flags = 0;		/* sends options! */
	tp->t_inpcb = inp;
	/*
	 * Init srtt to TCP_STV_SRTTBASE (0), so we can tell that we have no
	 * rtt estimate.  Set rttvar so that srtt + 2 * rttvar gives
	 * reasonable initial retransmit time.
	 */
	tp->t_srtt = TCP_STV_SRTTBASE;
	tp->t_rttvar = TCP_STV_SRTTDFLT << 2;
	TCP_ST_RANGESET(tp->t_rxtcur, 
	    ((TCP_STV_SRTTBASE >> 2) + (TCP_STV_SRTTDFLT << 2)) >> 1,
	    TCP_STV_MIN, TCP_STV_REXMTMAX);
	tp->snd_cwnd = sbspace(sototcp_sst(inp->inp_session)->snd);
	tp->snd_ssthresh = 65535;		/* XXX */
	inp->inp_ppcb = (caddr_t)tp;
	return (tp);
}

/*
 * Drop a TCP_S connection, reporting
 * the specified error.  If connection is synchronized,
 * then send a RST to peer.
 */
struct tcp_scb *
tcp_s_drop(tp, errnum)
	register struct tcp_scb *tp;
	int errnum;
{
	XObj so = tp->t_inpcb->inp_session;
	PSTATE	*ps = (PSTATE *)xMyProtl(so)->state;

	xTrace2(tcp_sp, 2, "tcp_s_drop: tcp_scb %X s %X", tp, so);
	if (TCP_SS_HAVERCVDSYN(tp->t_state)) {
	  tp->t_state = TCP_SS_CLOSED;
	  (void) tcp_s_output(tp);
	  ps->tcp_sstat->tcp_ss_drops++;
	} else {
	  ps->tcp_sstat->tcp_ss_conndrops++;
	}
	socantrcvmore(so);
#if 0
	return (tcp_s_destroy(tp));
#else
	return 0;
#endif
}

/*
 * Close a TCP_S control block:
 *	discard all space held by the tcp_s
 *	discard internet protocol block
 *	wake up any sleepers
 */
struct tcp_scb *
tcp_s_destroy(tp)
	register struct tcp_scb *tp;
{
	register struct reass *this, *next;
	struct inpcb *inp;
	XObj so;
	PSTATE *pstate;

	xTrace1(tcp_sp, 1, "tcp_s_destroy: tcp_scb %X", tp);
	inp = tp->t_inpcb;
	xTrace1(tcp_sp, 3, "tcp_s_destroy: inpcb %X", inp);
	so = inp->inp_session;
	pstate = (PSTATE *)so->myprotl->state;
	this = tp->seg_next;
	while (this != (struct reass *)tp) {
		next = this->next;
		msgDestroy(&this->m);
		remque(this);
		xFree((char *)this);
		this = next;
	}
	if (tp->t_template) {
	    tcp_sReleasePort(pstate->portstate, tp->t_template->ti_sport);
	    xFree((char *)tp->t_template);
	}
	xFree((char *)tp);
	inp->inp_ppcb = 0;
	/*
	 * This used to be a soisdisconnected, but really needs to delete
	 * all of the state.  
	 */
	delete_tcp_sstate(sototcp_sst(so));
	so->state = 0;
	mapRemoveBinding(pstate->activeMap, so->binding);
	{
		int i;

		for (i=0; i < so->numdown; i++) {
			xClose(xGetDown(so, i));
		}
	}
	xDestroy(so);

	in_pcbdetach(inp);
	pstate->tcp_sstat->tcp_ss_closed++;
	return ((struct tcp_scb *)0);
}


/*
 * Notify a tcp_s user of an asynchronous error;
 * just wake up so that he can collect error status.
 */
void
tcp_s_notify(inp)
	register struct inpcb *inp;
{

	sorwakeup(inp->inp_session);
	sowwakeup(inp->inp_session);
}


#if BSD<43
/* XXX fake routine */
tcp_s_abort(inp)
	struct inpcb *inp;
{
	return;
}
#endif

/*
 * When a source quench is received, close congestion window
 * to one segment.  We will gradually open it again as we proceed.
 */
void
tcp_s_quench(inp)
	struct inpcb *inp;
{
	struct tcp_scb *tp = intotcp_scb(inp);

	if (tp)
		tp->snd_cwnd = tp->t_maxseg;
}
